# Ubuntu CI Engine
# Copyright 2014 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import sys
import tempfile
import unittest

import mock

import lander.workflow


HERE = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(HERE, '../../bin'))
import lander_service_wrapper


class TestWrapper(unittest.TestCase):
    '''Sanity tests of the lander_service_wrapper.py script.'''

    def setUp(self):
        super(TestWrapper, self).setUp()
        patch = mock.patch('ci_utils.amqp_utils.get_config')
        patch.start()
        self.addCleanup(patch.stop)

        self.results = tempfile.mkstemp(prefix='lander-test-')[1]
        self.addCleanup(os.unlink, self.results)

        tsdata = {
            'id': 1,
            'steps': [
                {'name': 'ppa_creator'},
                {'name': 'bsbuilder'},
                {'name': 'image_builder'},
                {'name': 'test_runner'},
                {'name': 'publisher'},
            ]
        }
        workflow = lander.workflow.create('ts_url', tsdata)
        workflow.set_results('ppa_creator', {'ppa_name': 'ppa'})
        workflow.set_results('bsbuilder', {'ppa': 'ppa-name'})
        workflow.set_results('image_builder', {'image_id': 'id'})
        with open(self.results, 'w') as f:
            workflow.save(f)

        self.base_args = ['--results-file', self.results]

    def testBadService(self):
        '''make sure we exit gracefully for a bad service name'''
        args = self.base_args + ['--service', 'bad-name']
        with self.assertRaises(SystemExit):
            lander_service_wrapper._get_parser().parse_args(args)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testPPACreatorSucceeds(self, connection, send, api):
        '''Ensure we generate the proper response if things succeed.'''
        api().get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'uuid': 'AUUID',
        }
        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'PASSED',
                 'exit': True, 'name': 'ppa_creator'}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'ppa_creator']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(0, rc)

        api().set_ppacreation_waiting.assert_called_once()
        api().set_ppacreation_complete.assert_called_once_with(True)

        data = lander.workflow.load(self.results).get_step('ppa_creator')
        self.assertEqual(state, data)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testBSBuilderUnexpected(self, send, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api().get_full_ticket.return_value = {'subticket': 1}

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_pkg_waiting.assert_called_once()
        api().set_pkg_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('bsbuilder')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testBSBuilderSucceeds(self, connection, send, api):
        '''Ensure we generate the proper response if things succeed.'''

        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'PASSED',
                 'foo': 'bar', 'exit': True, 'name': 'bsbuilder'}
        api().get_full_ticket.return_value = {'subticket': 1}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'bsbuilder']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(0, rc)

        api().set_pkg_waiting.assert_called_once()
        api().set_pkg_complete.assert_called_once_with(True)

        data = lander.workflow.load(self.results).get_step('bsbuilder')
        self.assertEqual(state, data)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testImageBuilderUnexpected(self, send, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api().get_full_ticket.return_value = {
            'master_ppa': 1, 'series': 1, 'base_image': 1}

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_image_waiting.assert_called_once()
        api().set_image_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('image_builder')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testImageBuilderSucceeds(self, connection, send, api):
        '''Ensure we generate the proper response if things succeed.

        While technically "succeeding" its checking for a successful FAILURE
        '''
        api().get_full_ticket.return_value = {
            'master_ppa': 1, 'series': 1, 'base_image': 1}
        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'FAILED',
                 'foo': 'bar', 'exit': True, 'name': 'image_builder'}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'image_builder']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(1, rc)

        api().set_image_waiting.assert_called_once()
        api().set_image_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('image_builder')
        self.assertEqual(state, data)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testPublisherUnexpected(self, send, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api().get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',
            'subticket': [
                {'id': 1,
                 'source_package_upload': {
                     'sourcepackage': {'name': 'foobar'}}}],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

        args = self.base_args + ['--service', 'publisher']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_publishing_waiting.assert_called_once()
        api().set_publishing_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('publisher')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testPublisherSucceeds(self, connection, send, api):
        '''Ensure we generate the proper response if things succeed.

        While technically "succeeding" its checking for a successful FAILURE
        '''
        api().get_full_ticket.return_value = {
            'ticket_id': 1,
            'progress_trigger': 'A_TRIGGER',

            'subticket': [
                {'id': 1,
                 'source_package_upload': {
                     'sourcepackage': {'name': 'foobar'}}}],
            'source_archive': 'A_PPA',
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'current_workflow_step': 'Queued',
        }

        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'FAILED',
                 'foo': 'bar', 'exit': True, 'name': 'publisher'}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'publisher']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(1, rc)

        api().set_publishing_waiting.assert_called_once()
        api().set_publishing_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('publisher')
        self.assertEqual(state, data)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    @mock.patch('ci_utils.amqp_utils.connection')
    def testRunnerSucceeds(self, connection, send, api):
        '''Ensure we generate the proper response if things succeed.'''
        api().get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'subticket': [
                {'id': 1,
                 'source_package_upload': {
                     'sourcepackage': {'name': 'foobar'}}}],
        }
        # mock the run-forever logic to ensure we return what we expect
        state = {'state': 'COMPLETED', 'result': 'PASSED',
                 'foo': 'bar', 'exit': True, 'name': 'test_runner'}

        def amqp():
            def run_forever(channel, queue, callback):
                m = mock.Mock()
                m.body = json.dumps({'state': 'WAITING'})
                callback(m)
                m.body = json.dumps(state)
                callback(m)
            return run_forever

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)

        with mock.patch('ci_utils.amqp_utils._run_forever', new_callable=amqp):
            rc = lander_service_wrapper.main(args)
            self.assertEqual(0, rc)

        api().set_testing_waiting.assert_called_once()
        api().set_testing_complete.assert_called_once_with(True)

        data = lander.workflow.load(self.results).get_step('test_runner')
        self.assertEqual(state, data)

    @mock.patch('lander.ticket_api.TicketApi')
    @mock.patch('ci_utils.amqp_utils.send')
    def testRunnerUnexpected(self, send, api):
        '''Ensure unexpected errors still generate a response'''
        err_msg = 'foo bar'
        send.side_effect = RuntimeError(err_msg)
        api().get_full_ticket.return_value = {
            'master_ppa': 'ANOTHER_PPA',
            'series': 'A_SERIES',
            'subticket': [
                {'id': 1,
                 'source_package_upload': {
                     'sourcepackage': {'name': 'foobar'}}}],
        }

        args = self.base_args + ['--service', 'test_runner']
        args = lander_service_wrapper._get_parser().parse_args(args)
        rc = lander_service_wrapper.main(args)
        self.assertEqual(1, rc)
        api().set_testing_waiting.assert_called_once()
        api().set_testing_complete.assert_called_once_with(False)

        data = lander.workflow.load(self.results).get_step('test_runner')
        self.assertEqual('FAILED', data['result'])
        self.assertEqual(err_msg, data['error'])
