var app = angular.module('NonFunctional', ['ngAnimate', 'nvd3ChartDirectives', 'ngQuickDate']);

var DOMAIN = '';
var API_PREFIX = "/api/v1/";
var GRAPH_DIR = "graphs/";
var ISO_EIGHT_SIX_OH_ONE = "%Y-%m-%dT%H:%M:%SZ";
var ISO_ISH = "%Y-%m-%d %H:%M:%S %Z";

if (window.location.origin == 'file://') {
    DOMAIN = 'http://nfss.staging.ubuntu.com';
}

// This is just a basic data table to use in the event that a given test or
// project does not have a real graph definition yet.
var DEFAULT_GRAPH = '<ol><li ng-repeat="item in graphData[0].values">' +
    '<b>{{dateFormatter()(dateParser(item.date_entered))}}:</b>' +
    '<br />{{item.data}}</li></ol>';

// This allows reversing the display order of lists in angular templates
// by doing something like ng-repeat="foo in bar | reverse"
app.filter('reverse', function() {
    return function(items) {
        return items.slice().reverse();
    };
});

// 'item' refers to each entry in the blob.data array, and allows for
// graph definitions that look like y="accessor('delta')"
function accessor(key) { return function(item) { return item[key]; } };

// Produce a date range for the REST API if appropriate
function dateRange(url) {
    var start = url.start_date;
    var end =   url.end_date;
    if (start && end) {
        return '?start_date=' + start + '&end_date=' + end;
    }
    return '';
}

// Convert "2014-04-15 09:35:07.926+00" into a legit JS date object.
// Note, it's "parts[1] - 1" because months are 0-indexed.
function dateParser(dateString) {
    var parts = dateString.split(/[-:. TZ]/g);
    return new Date(Date.UTC(parts[0], parts[1] - 1, parts[2],
                             parts[3], parts[4], parts[5]));
}

// Display dates nicely. For information on the format string, refer to
// https://github.com/mbostock/d3/wiki/Time-Formatting#format
function dateFormatter(formatString) {
    return function(date) {
        return d3.time.format(formatString || ISO_ISH)(new Date(date));
    }
}

// Display numbers nicely. For information on the format string, refer to
// https://github.com/mbostock/d3/wiki/Formatting#d3_format
function numberFormatter(formatString, units) {
    return function(number) {
        return (d3.format(formatString || ',.2f')(number) + (units || ''));
    }
}

// Create a new date object that is an arbitrary number of months offset
// from the passed-in date object.
function deltaMonth(date, months) {
    date = new Date(date);
    date.setMonth(date.getMonth() + months);
    return date;
}

// Generate a URL query string that contains the appropriate date range
// defined by the arguments start and end (which should be date objects).
function makeMonthLink(url, start, end) {
    return ('#?project=' + url.project + '&test=' + url.test +
            '&start_date=' + (start.getTime() / 1000) + '&end_date=' + (end.getTime() / 1000));
}

// Return a string that identifies the current chosen project + test.
function makeSlug(URLGetter) {
    return function() {
        var url = URLGetter();
        if (url.project && url.test) {
            return url.project + '_' + url.test;
        }
    }
}

function appController($scope, $location, $http) {
    $scope.loading = false;

    $scope.endDate = null;
    $scope.startDate = null;

    // Expose some global funcs to the template.
    $scope.accessor = accessor;
    $scope.dateParser = dateParser;
    $scope.dateFormatter = dateFormatter;
    $scope.numberFormatter = numberFormatter;

    function dateWatcher() {
        if ($scope.startDate && $scope.endDate) {
            $scope.url.start_date = $scope.startDate.getTime() / 1000;
            $scope.url.end_date   = $scope.endDate.getTime()   / 1000;
            $location.search($scope.url);
            $scope.getGraphData();
        }
    }

    $scope.$watch('endDate', dateWatcher);
    $scope.$watch('startDate', dateWatcher);

    $scope.ISO_ISH = ISO_ISH;

    // Parse URL query paramters into a dict and return it.
    function URLParams() {
        return $location.search();
    }

    // Watch for changes in the URL and apply that data to the scope.
    // Values are stored as eg $scope.url.project
    function URLWatcher(queryParams) {
        $scope.url = queryParams;
    }

    function URLGetter() {
        return $scope.url;
    }

    $scope.$watch(URLParams, URLWatcher);

    function setChosenProjectData(data) {
        $scope.chosenProjectData = data.tests;
    }

    // When a project is chosen, get a list of its tests
    function projectWatcher(chosenProject) {
        $scope.tests = [];
        $scope.graphData = [];
        if (chosenProject in $scope.allProjects) {
            $scope.tests = $scope.allProjects[chosenProject].tests.sort();
            $http.get(DOMAIN + $scope.allProjects[chosenProject].path)
                 .success(setChosenProjectData);
        }
    }

    // Handle results from initial rest request: Show the list of projects.
    function projectInitializer(data) {
        $scope.projects = Object.keys(data.projects).sort();
        $scope.allProjects = data.projects;
        $scope.$watch('url.project', projectWatcher);
    }

    // Make the initial rest request.
    $http.get(DOMAIN + API_PREFIX).success(projectInitializer);

    $scope.slug = makeSlug(URLGetter);

    // d3 expects data in a certain format; this is the most basic possible
    // format that makes no assumptions about the structure of the data coming
    // from the DB. Graph definitions have the ability to override this.
    $scope.defaultMassageGraphData = function(blob) {
        $scope.graphData = [ { values: blob.data } ];
    }

    // Set the values for the next/previous month links after the graph data has loaded.
    function setMonthLinks(blob) {
        $scope.loading = false;
        $scope.hidePrevLink = false;
        $scope.hideNextLink = false;
        $scope.veryFirst = dateParser(blob.first_data);
        $scope.veryLast = dateParser(blob.last_data);
        var first = $scope.url.start_date ? new Date($scope.url.start_date * 1000) : dateParser(blob.data_range_first);
        var last  = $scope.url.end_date   ? new Date($scope.url.end_date   * 1000) : dateParser(blob.data_range_last);
        if (first.getTime() <= $scope.veryFirst.getTime()) {
            $scope.hidePrevLink = true;
        }
        if (last.getTime() >= $scope.veryLast.getTime()) {
            $scope.hideNextLink = true;
        }
        $scope.prevMonthLink = makeMonthLink($scope.url, deltaMonth(first, -1), first);
        $scope.nextMonthLink = makeMonthLink($scope.url, last, deltaMonth(last, 1));
    }

    // Fetch the actual graph data from the db.
    $scope.getGraphData = function() {
        $scope.loading = true;
        $scope.graphData = [];
        $http.get(DOMAIN + API_PREFIX + $scope.url.project + '/' + $scope.url.test + dateRange($scope.url))
        .success(setMonthLinks)
        .success($scope.massageGraphData);
    }
}

app.directive('customGraph', function($compile, $http, $templateCache) {
    // This function handles the switching of graph definitions and loading of
    // graph data when the user navigates between pages. The functions defined
    // within this function will make more sense if you read them bottom to top.
    function graphLinker(scope, element, attrs) {

        // Take a string of raw HTML data, insert it into the DOM, and
        // tell angular that the DOM has changed so it can make it go.
        function graphChanger(htmlData) {
            element.html(htmlData || DEFAULT_GRAPH);
            $compile(element.contents())(scope);
            var script = element.contents()[0];
            if (script.tagName == "SCRIPT") {
                // script.text should contain a meaningful definition of
                // scope.massageGraphData
                eval(script.text);
            }
            // Now that the graph definition is ready, let's load the graph data.
            scope.getGraphData();
        }

        // Make the AJAX request to load the raw HTML data, calling graphChanger
        // on success, returns a promise so you can handle errors.
        function getGraphDef(slug) {
            scope.loading = true;
            return $http({
                method: 'GET',
                url: GRAPH_DIR + slug + '.html',
                cache: $templateCache
            }).success(graphChanger);
        }

        // When all else fails, display the raw data in a list.
        function setGlobalDefaultGraph() {
            graphChanger(DEFAULT_GRAPH);
        }

        // When there is neither a project+test specific graph definition, nor
        // a generic test-only graph definition, try looking for a generic
        // project-only graph definition. Eg, this loads bootspeed.html
        function setProjectDefaultGraph() {
            getGraphDef(scope.url.project).error(setGlobalDefaultGraph);
        }

        // When there is no project+test specific graph definition, look
        // instead of a generic test-only graph definition which serves as a
        // default for all projects that run that test. Eg, app_startup_benchmark.html
        function setTestDefaultGraph() {
            getGraphDef(scope.url.test).error(setProjectDefaultGraph);
        }

        // Attempt to load a project+test specific graph definition, this
        // can be used to override the graph display on a case-by-case basis.
        scope.graphWatcher = function(urlParams) {
            scope.graphData = [];
            scope.startDate = null;
            scope.endDate = null;
            var slug = scope.slug();
            if (slug) {
                scope.massageGraphData = scope.defaultMassageGraphData;
                getGraphDef(slug).error(setTestDefaultGraph);
            }
        };

        scope.$watch('url', scope.graphWatcher);
    }

    return {
        replace: true,
        link: graphLinker
    };
});
