describe('The accessor function', function() {
    it('should return the correct values.', function() {
        var data = { foo: 42, bar: 27 };
        expect(accessor('foo')(data)).toBe(42);
        expect(accessor('bar')(data)).toBe(27);
        expect(accessor('qux')(data)).toBeUndefined();
    });
});

describe('The dateRange function', function() {
    it('should return URL query parameters when appropriate.', function() {
        expect(dateRange({start_date: 1, end_date: 100}))
        .toBe('?start_date=1&end_date=100');
    });
    it('should not return URL query parameters otherwise.', function() {
        expect(dateRange({})).toBe('');
        expect(dateRange({start_date: 1000})).toBe('');
        expect(dateRange({end_date: 42})).toBe('');
    });
});

describe('The dateParser function', function() {
    it('should parse dates properly.', function() {
        var result = dateParser('2014-01-02T03:04:05Z');
        expect(result instanceof Date).toBe(true);
        expect(result.getUTCFullYear()).toBe(2014);
        expect(result.getUTCMonth()).toBe(0);
        expect(result.getUTCDate()).toBe(2);
        expect(result.getUTCHours()).toBe(3);
        expect(result.getUTCMinutes()).toBe(4);
        expect(result.getUTCSeconds()).toBe(5);

        var result = dateParser('2008-08-30 12:30:45');
        expect(result.toUTCString()).toBe('Sat, 30 Aug 2008 12:30:45 GMT');
    });
});

describe('The dateFormatter function', function() {
    it('should use a sensible default format.', function() {
        var now = new Date();
        spyOn(d3.time, 'format').and.returnValue(
            function() { return arguments });
        var result = dateFormatter()(now);
        expect(d3.time.format).toHaveBeenCalledWith('%Y-%m-%d %H:%M:%S %Z');
        expect(result[0]).toEqual(now);
    });
    it('should accept a custom format.', function() {
        var now = new Date();
        spyOn(d3.time, 'format').and.returnValue(
            function() { return arguments });
        var result = dateFormatter('%Y-%m-%d')(now);
        expect(d3.time.format).toHaveBeenCalledWith('%Y-%m-%d');
        expect(result[0]).toEqual(now);
    });
});

describe('The numberFormatter function', function() {
    it('should use a sensible default format.', function() {
        var n = 49.3;
        spyOn(d3, 'format').and.returnValue(
            function() { return arguments[0] });
        var result = numberFormatter(null, ' parsecs')(n);
        expect(d3.format).toHaveBeenCalledWith(',.2f');
        // Unformatted because d3 is mocked out.
        expect(result).toEqual('49.3 parsecs');
    });
    it('should accept a custom format.', function() {
        var n = 24.7;
        spyOn(d3, 'format').and.returnValue(
            function(n) { return n });
        var result = numberFormatter('.6f', ' seconds')(n);
        expect(d3.format).toHaveBeenCalledWith('.6f');
        // Unformatted because d3 is mocked out.
        expect(result).toEqual('24.7 seconds');
    });
    it('should not require a unit.', function() {
        var n = 24.123;
        spyOn(d3, 'format').and.returnValue(
            function(n) { return n });
        expect(numberFormatter('.2f')(n)).toBe('24.123');
    });
});

describe('The deltaMonth function', function() {
    it('should be able to offset months.', function() {
        var result = deltaMonth(new Date(2014, 0, 15), -1);
        expect(result.getMonth()).toBe(11);
        expect(result.getYear()).toBe(113);
        var result = deltaMonth(new Date(2014, 0, 15), 1);
        expect(result.getMonth()).toBe(1);
        expect(result.getYear()).toBe(114);
    });
});

describe('The makeMonthLink function', function() {
    it('should return URL query parameters.', function() {
        var url = { project: 'foo', test: 'bar' };
        var start = new Date(Date.UTC(1970, 0, 1));
        var end   = new Date(Date.UTC(1970, 0, 2));
        expect(makeMonthLink(url, start, end))
        .toBe('#?project=foo&test=bar&start_date=0&end_date=86400');
    });
});

describe('The makeSlug function', function() {
    it('should make a slug.', function() {
        var url = {};
        function URLGetter() { return url; };
        var slugger = makeSlug(URLGetter);
        expect(slugger()).toBeUndefined();
        url.project = 'foo';
        expect(slugger()).toBeUndefined();
        url.test = 'bar';
        expect(slugger()).toBe('foo_bar');
        delete url.project;
        expect(slugger()).toBeUndefined();
    });
});

describe('The appController function', function() {
    var scopeMock, locationMock, httpMock;
    beforeEach(function() {
        scopeMock = {
            'url': {},
            '$watch': NOP,
        };
        locationMock = {
            'search': NOP,
        };
        httpMock = {
            'get': function () {
                return {
                    success: function() { return this },
                    error: function() { return this }
                }
            }
        };
    });
    it('should set some basic state.', function() {
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.loading).toBe(false);
        expect(scopeMock.endDate).toBeNull();
        expect(scopeMock.startDate).toBeNull();
        expect(scopeMock.accessor).toBe(accessor);
        expect(scopeMock.dateParser).toBe(dateParser);
        expect(scopeMock.dateFormatter).toBe(dateFormatter);
        expect(scopeMock.numberFormatter).toBe(numberFormatter);
        expect(scopeMock.slug).toBeDefined();
        expect(scopeMock.ISO_ISH).toBe('%Y-%m-%d %H:%M:%S %Z');
    });
    it('should watch for changes.', function() {
        spyOn(scopeMock, '$watch');
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var watchCalls = scopeMock.$watch.calls.all();
        expect(watchCalls[0].args[0]).toBe('endDate');
        expect(watchCalls[0].args[1].name).toBe('dateWatcher');
        expect(watchCalls[1].args[0]).toBe('startDate');
        expect(watchCalls[1].args[1].name).toBe('dateWatcher');
        expect(watchCalls[2].args[0].name).toBe('URLParams');
        expect(watchCalls[2].args[1].name).toBe('URLWatcher');
    });
    it('should respond to the date picker.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search');
        appController(scopeMock, locationMock, httpMock);
        spyOn(scopeMock, 'getGraphData');
        var dateWatcher = scopeMock.$watch.calls.first().args[1];
        dateWatcher();
        expect(scopeMock.url.start_date).toBeUndefined();
        expect(scopeMock.url.end_date).toBeUndefined();
        scopeMock.startDate = new Date();
        dateWatcher();
        expect(scopeMock.url.start_date).toBeUndefined();
        expect(scopeMock.url.end_date).toBeUndefined();
        scopeMock.endDate = new Date();
        dateWatcher();
        expect(scopeMock.url.start_date).toBe(scopeMock.startDate.getTime() / 1000);
        expect(scopeMock.url.end_date).toBe(scopeMock.endDate.getTime() / 1000);
        expect(locationMock.search).toHaveBeenCalledWith(scopeMock.url);
        expect(scopeMock.getGraphData).toHaveBeenCalledWith();
    });
    it('should have a URLParams function.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search').and.returnValue('sentinel');
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var URLParams = scopeMock.$watch.calls.all()[2].args[0];
        var result = URLParams();
        expect(locationMock.search).toHaveBeenCalled();
        expect(result).toBe('sentinel');
    });
    it('should have a URLWatcher function.', function() {
        spyOn(scopeMock, '$watch');
        spyOn(locationMock, 'search').and.returnValue('sentinel');
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.$watch).toHaveBeenCalled();
        var URLWatcher = scopeMock.$watch.calls.all()[2].args[1];
        URLWatcher('hiya!');
        expect(scopeMock.url).toBe('hiya!');
    });
    it('should have a projectInitializer and projectWatcher.', function() {
        var projectInitializer;
        spyOn(scopeMock, '$watch');
        spyOn(httpMock, 'get').and.returnValue({
            success: function(f) {
                projectInitializer = f;
            }
        });
        appController(scopeMock, locationMock, httpMock);
        expect(httpMock.get).toHaveBeenCalled();
        var data = { projects: { foo: 1, bar: 2, baz: 3 } };
        projectInitializer(data);
        expect(scopeMock.projects).toEqual([ 'bar', 'baz', 'foo' ]);
        expect(scopeMock.allProjects).toBe(data.projects);
        expect(scopeMock.$watch).toHaveBeenCalled();
        watchCall = scopeMock.$watch.calls.mostRecent();
        expect(watchCall.args[0]).toBe('url.project');
        var projectWatcher = watchCall.args[1];
        expect(projectWatcher.name).toBe('projectWatcher');
        scopeMock.allProjects = {
            foo: {
                tests: [ 'one', 'two', 'three' ],
                path: '/api/path/here',
            },
        };
        scopeMock.tests = 'wrong';
        scopeMock.graphData = 'bad';
        expect(httpMock.get.calls.count()).toBe(1);
        projectWatcher('foo');
        expect(scopeMock.tests).toEqual([ 'one', 'three', 'two' ]);
        expect(scopeMock.graphData).toEqual([]);
        expect(httpMock.get.calls.mostRecent().args[0]).toBe(DOMAIN + '/api/path/here');
        expect(httpMock.get.calls.count()).toBe(2);
        projectWatcher('nonexistent');
        expect(scopeMock.tests).toEqual([]);
        expect(scopeMock.graphData).toEqual([]);
        expect(httpMock.get.calls.count()).toBe(2);
    });
    it('should have a default way of massaging graph data.', function() {
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.defaultMassageGraphData).toBeDefined();
        var blob = {
            irrelevant: 'stuff',
            data: 'oh yeah'
        };
        scopeMock.defaultMassageGraphData(blob);
        expect(scopeMock.graphData).toBeDefined();
        expect(scopeMock.graphData.length).toBe(1);
        expect(scopeMock.graphData[0].values).toBe('oh yeah');
    });
    it('should have a getGraphData function.', function() {
        appController(scopeMock, locationMock, httpMock);
        expect(scopeMock.getGraphData).toBeDefined();
        scopeMock.url = {
            project: 'foo',
            test: 'bar',
        };
        spyOn(httpMock, 'get').and.callThrough();
        expect(scopeMock.loading).toBe(false);
        scopeMock.getGraphData();
        expect(scopeMock.loading).toBe(true);
        expect(scopeMock.graphData).toEqual([]);
        expect(httpMock.get).toHaveBeenCalledWith(DOMAIN + API_PREFIX + 'foo/bar');
        scopeMock.url.start_date = 1;
        scopeMock.url.end_date = 69;
        scopeMock.getGraphData();
        expect(httpMock.get).toHaveBeenCalledWith(
            DOMAIN + API_PREFIX +
            'foo/bar?start_date=1&end_date=69');
    });
    it('should know when to set month links.', function() {
        var successArgs = [];
        scopeMock.massageGraphData = 'i need a massage.';
        appController(scopeMock, locationMock, httpMock);
        spyOn(httpMock, 'get').and.returnValue({
            success: function(f) {
                successArgs.push(f);
                return this;
            }
        });
        expect(scopeMock.loading).toBe(false);
        scopeMock.getGraphData();
        expect(scopeMock.loading).toBe(true);
        var setMonthLinks = successArgs[0];
        expect(successArgs[1]).toBe('i need a massage.');
        scopeMock.url = {
            project: 'flib',
            test: 'jib',
        };
        var blob = {
            data_range_first: "2014 01 01 12 59 59",
            data_range_last: "2014 02 02 15 59 59",
        };
        blob.first_data = blob.data_range_first;
        blob.last_data  = blob.data_range_last;
        setMonthLinks(blob);
        expect(scopeMock.loading).toBe(false);
        expect(scopeMock.veryFirst instanceof Date).toBe(true);
        expect(scopeMock.veryLast instanceof Date).toBe(true);
        expect(scopeMock.hidePrevLink).toBe(true);
        expect(scopeMock.hideNextLink).toBe(true);
        expect(scopeMock.prevMonthLink).toBe(
            '#?project=flib&test=jib&start_date=1385902799&end_date=1388581199');
        expect(scopeMock.nextMonthLink).toBe(
            '#?project=flib&test=jib&start_date=1391356799&end_date=1393775999');
        blob.first_data = '1970-10-20T23:11:30Z';
        setMonthLinks(blob);
        expect(scopeMock.hidePrevLink).toBe(false);
        expect(scopeMock.hideNextLink).toBe(true);
        blob.last_data = '2015-12-25 12:13:14';
        setMonthLinks(blob);
        expect(scopeMock.hidePrevLink).toBe(false);
        expect(scopeMock.hideNextLink).toBe(false);
    });
});
